# This text include all R functions for implementing SC, gSC(m0), and FR on high dimensional quantile regression models.
#
# Written by
# Chien-Tong Lin (ctlin@mx.nthu.edu.tw) 
# Taiwan, Aug. 2021
# 
# 
# List of R functions:
# Part I: main functions
#   1) rq.margin: main function for calculating the (conditional) marginal utility
#   2) rq.greedy: main function for implementing greedy algorithms
# Part II: additional functions
#   1) standardize:  standardize each covariate

###########################################################################
# Part I: main functions
# Part I 1) rq.margin: main function for calculating the (conditional) marginal utility
#
# Input:
#   z: n x p design matrix
#   y: the response 
#   tau: the quantile level
#   ind.act: the conditioning set
#   ind.cand: the candidate set
#   intercept: whether or not to include the intercept in the model (TRUE or FALSE)
#
# Output: 
#   score: the score for the marignal utility (condition on the ind.act)


rq.margin <- function(z, y, tau, ind.act=NULL, ind.cand=NULL, intercept=TRUE, method="br"){
  
  z <- as.matrix(z)
  n <- nrow(z)
  p <- ncol(z)
  if(is.null(ind.cand)){
    ind.cand <- setdiff(1:p, ind.act)  
  }
  n.ind.act <- length(ind.act)
  n.ind.cand <- length(ind.cand)
  score.m <- rep(NA, p)
  
  ### Marginal regression
  for(i in 1:n.ind.cand){
    ind <- sort(c(as.numeric(ind.act), as.numeric(ind.cand[i])))
    if(intercept){
      obj <- rq(y ~ as.matrix(z[,ind]), tau=tau, method=method)  
    } else {
      obj <- rq(y ~ as.matrix(z[,ind]) - 1, tau=tau, method=method)  
    }
    score.m[ind.cand[i]] <- mean(check(obj$residuals, tau))
    if(is.infinite(abs(score.m[ind.cand[i]]))){
      score.m[ind.cand[i]] <- NA
    }
  }
  
  return(score=score.m)
}
###########################################################################
# Part I: main functions
# Part I 2) rq.greedy: main function for generating the greedy path
#
# Input:
#   x: n x p design matrix
#   y: the response 
#   tau: the quantile level
#   screen.method: choose the method of greedy algorithm (SC, gSC, FR)
#   stop.method: top the greedy algorithm by LIC ( or set ``none'' to stop until k hits max.iter)
#   intercept: whether or not to include the intercept in the model (TRUE or FALSE)
#   xi: the tuning parameter for LIC (xi in the paper)
#   m: the parameter for gSC (m0 in the paper)
#   K: stop the algorithm when LIC is satisfies for K times (T in the paper)
#   max.iter: the maximum parameter (Kn in the paper)
# Output: 
#   beta: the solution path generated by the greedy algorithm
#   loss: the loss corresponds to the solution path beta
#   ind.act: the index set selected by the greedy algorithm
#   LIC: a sequence of LIC values corresponds to the greedy path
#   HQIC: a sequence of HQIC values corresponds to the greedy path
#   runtime: computing time

rq.greedy <- function(x, y, tau, screen.method, stop.method="none", xi=1, m=50, K=3, max.iter=30, method="br", c0=1){

  x <- as.matrix(x)
  n <- nrow(x); p <- ncol(x)
  obj <- rq(y ~ 1, tau=tau)
  Res <- obj$residuals
  
  time.start <- Sys.time()

  ### initial setting
  info.vec <- rep(NA, max.iter)
  beta.path <- matrix(0, nrow=p+1, ncol=max.iter)
  rownames(beta.path) <- c("intercept", 1:p)
  ind.LIC <- ind.QIC <- ind.act <- NULL
  ind.list <- vector("list", max.iter)
  ind.cand <- setdiff(1:p, ind.act)  
  LIC.vec <- HQIC.vec <- df <- loss.vec <- rep(NA, max.iter)
  k0 <- 0

  for(i in 1:max.iter){

    ### variable screening
    if(screen.method=="SC"){
      score <- rq.margin(x, Res, tau=tau, ind.act=NULL, intercept=FALSE, method=method)
    } else if (screen.method=="FR"){
      score <- rq.margin(x, y, tau=tau, ind.act=ind.act, method=method)
    } else if (screen.method=="PQU.fr"){
      score <- -Screen_Mar(x, Res, tt=tau, 1)
    } else if (screen.method=="CGA"){
      grad <- crossprod(x, tau - ifelse(Res<=0, 1, 0))/n
      score <- -abs(grad)
    } else if (screen.method=="SC+FR"){
      s0 <- rq.margin(x, Res, tau=tau, ind.act=NULL, ind.cand=ind.cand, intercept=FALSE, method=method)
      s0[ind.act] <- Inf
      ind <- which((p-rank(-s0, ties.method="max")+1)<=m)
      score <- rq.margin(x, y, tau=tau, ind.act=ind.act, ind.cand=ind, method=method)
    } else if (screen.method=="CGA+FR"){
      grad <- crossprod(x, tau - ifelse(Res<=0, 1, 0))/n
      s0 <- -abs(grad); s0[ind.act] <- Inf
      ind <- which((p-rank(-s0, ties.method="max")+1)<=m)
      score <- rq.margin(x, y, tau=tau, ind.act=ind.act, ind.cand=ind, method=method)
    }
    score[ind.act] <- Inf
    ind.act <- c(ind.act, which.min(score))
    ind.cand <- setdiff(1:p, ind.act)  
    df[i] <- length(ind.act)

    ### refit MLE (with intercept)
    obj <- rq(y ~ as.matrix(x[,ind.act]), tau=tau, method=method)
    Res <- obj$residuals
    loss.vec[i] <- mean(check(Res, tau))
    beta.path[c(1,ind.act+1), i] <- obj$coefficients
    
    if(is.nan(loss.vec[i])|is.infinite(abs(loss.vec[i]))){
      i <- i-1;
      ind.act <- ind.act[1:i]
      break;
    }
    
    ### recording information criterion
    LIC.vec[i] <- loss.vec[i] + (df[i]^2)*xi*log(p)/n
    HQIC.vec[i] <- log(loss.vec[i]*n) + c0*df[i]*log(n)*log(log(p))/n/2
    
    ### stopping criterion
    if(i>1){
      if(stop.method=="LIC"){
        if(loss.vec[i-1]-loss.vec[i] < xi*(i-1)*log(p)/n){
          k0 <- k0 + 1
        } else {
          k0 <- 0
        }
      } 
    }

    # cat("i=", i, "; loss=", loss.vec[i], '\n')
    # print(ind.act)

    if(k0==K) break;
    if(i==max.iter) break;
  }
  
  # beta.path <- sweep(beta.path, 1, normx, "/")
  
  time.end <- Sys.time()
  runtime <- difftime(time.end, time.start, units="secs")
  
  list(beta=beta.path[,1:i], 
       loss=loss.vec[1:i],
       ind.act=ind.act,
       LIC=LIC.vec[1:i],
       HQIC=HQIC.vec[1:i],
       runtime=runtime)
}

###########################################################################
# Part II: additional funcitons
# Part II 1) standardize: standardize each of p covariates
#
# Input:
#   X:   n x p design matrix
#
# Output: 
#   x:          (standardized) n x p design matrix
#   unit.var:   the scaling factors for columns of X

standardize = function (X)
{
  # X <- as.matrix(X)
  n <- nrow(X)
  p <- ncol(X)
  
  if(p>1){
    center = colMeans(X)
    X.c = sweep(X, 2, center)
    unit.var = sqrt(apply(X.c, 2, crossprod))
    # val = sweep(X.c, 2, unit.var/sqrt(n), "/")
    val <- sweep(X.c, 2, unit.var, "/")
  } else {
    X.c = X - mean(X)
    unit.var = sqrt(crossprod(X.c))
    # val = X.c/as.numeric(sqrt(crossprod(X.c))/sqrt(n))
    val = X.c/as.numeric(sqrt(crossprod(X.c)))
  }
  
  # list(x=val, unit.var=unit.var/sqrt(n))
  list(x=val, unit.var=unit.var)
}