# This text include all R functions for applying methods to the gene expression data from Scheetz et al. (2006)
#
# The data can be found in the package R::flare
# 
# Written by
# Chien-Tong Lin (ctlin@mx.nthu.edu.tw) 
# Taiwan, Aug. 2021
# 
# 
# List of R functions:
# Part I: main functions
#   1) RDA: implement all methods of the paper to select relevant genes in real data anlaysis (RDA)
#   2) RDA.pred: evaluate the prediction error (PE) off all methods of the paper
#
# Part II: additional functions
#   1) rq.margin: To implement (conditional) marginal utilities.
#   2) rq.greedy: To implement greedy algorithms.
#   3) QICDpen: To implement SCAD and MCP for quantile regression models.
#   4) Screen_Mar: Marginal screening as in Biometrika 2014.
#
###########################################################################


###########################################################################
# Part I: main functions
# Part I 1) RDA: implement all methods of the paper to select relevant genes in real data anlaysis (RDA)
#
# Input:
#   x:  n x p design matrix
#   y: the response vector
#   tau: the quantile level
#
# Output: 
#   the indice sets and corresponding regression estimates with respect to all methods considered in the paper

RDA <- function(x, y, tau, c0=1, seed=NULL){
  
  
  n <- nrow(x); p <- ncol(x)
  m <- ceiling(n/log(n))
  xi <- log(log(n))
  c0 <- log(log(p))/log(p)
  
  ### LASSO
  obj.lasso <- QICD.pen2(x=x, y=y, tau=tau, penalty="LASSO", c0=1/log(n)/log(p))
  
  pf <- 1/(abs(obj.lasso$beta.hqic) + 1e-5)
  obj.alasso <- ALASSO(x=x, y=y, tau=tau, ind=obj.lasso$ind.hqic, penalty.factor=pf, c0=c0)
  
  obj.sc <- rq.greedy(x=x, y=y, tau=tau, screen.method="SC", stop.method="LIC", xi=xi, method="br")
  obj.sc.scad <- QICD.pen2(x=x, y=y, tau=tau, penalty="SCAD", ind=obj.sc$ind.act, c0=c0)
  obj.sc.mcp <- QICD.pen2(x=x, y=y, tau=tau, penalty="MCP", ind=obj.sc$ind.act, c0=c0)
  
  obj.gsc <- rq.greedy(x=x, y=y, tau=tau, screen.method="SC+FR", stop.method="LIC", xi=xi, method="br", m=m)
  obj.gsc.scad <- QICD.pen2(x=x, y=y, tau=tau, penalty="SCAD", ind=obj.gsc$ind.act, c0=c0)
  obj.gsc.mcp <- QICD.pen2(x=x, y=y, tau=tau, penalty="MCP", ind=obj.gsc$ind.act, c0=c0)
  
  obj.gsch <- rq.greedy.h(x=x, y=y, tau=tau, stop.method="LIC", xi=xi, method="br")
  obj.gsch.scad <- QICD.pen2(x=x, y=y, tau=tau, penalty="SCAD", ind=obj.gsch$ind.act, c0=c0)
  obj.gsch.mcp <- QICD.pen2(x=x, y=y, tau=tau, penalty="MCP", ind=obj.gsch$ind.act, c0=c0)
  
  
  obj.fr <- rq.greedy(x=x, y=y, tau=tau, screen.method="FR", stop.method="LIC", xi=xi, method="br")
  obj.fr.scad <- QICD.pen2(x=x, y=y, tau=tau, penalty="SCAD", ind=obj.fr$ind.act, c0=c0)
  obj.fr.mcp <- QICD.pen2(x=x, y=y, tau=tau, penalty="MCP", ind=obj.fr$ind.act, c0=c0)
  
  ### PQ
  obj.pq <- which(Screen_Mar(x, y, tt=tau, ceiling(n/log(n)))==1)
  
  ### PQ+SCAD
  obj.pq.scad <- QICD.pen2(x=x, y=y, tau=tau, penalty="SCAD", ind=obj.pq, c0=c0)
  obj.pq.mcp <- QICD.pen2(x=x, y=y, tau=tau, penalty="MCP", ind=obj.pq, c0=c0)
  
  ### PQFR
  obj.pqfr <- rq.greedy(x, y, tau=tau, screen.method="PQU.fr", c0=1)
  ind.pqfr <- obj.pqfr$ind.act[which.min(obj.pqfr$HQIC)]
  obj.pqfr.scad <- QICD.pen2(x=x, y=y, tau=tau, penalty="SCAD", ind=ind.pqfr, c0=c0)
  obj.pqfr.mcp <- QICD.pen2(x=x, y=y, tau=tau, penalty="MCP", ind=ind.pqfr, c0=c0)
  
  
  list(obj.pq    =obj.pq,
       obj.pq.scad    =obj.pq.scad,
       obj.pq.mcp    =obj.pq.mcp,
       obj.sc    =obj.sc,
       obj.sc.scad = obj.sc.scad,
       obj.sc.mcp = obj.sc.mcp,
       obj.gsc    =obj.gsc,
       obj.gsc.scad = obj.gsc.scad,
       obj.gsc.mcp = obj.gsc.mcp,
       obj.gsch    =obj.gsch,
       obj.gsch.scad = obj.gsch.scad,
       obj.gsch.mcp = obj.gsch.mcp,
       obj.fr    =obj.fr,
       obj.fr.scad = obj.fr.scad,
       obj.fr.mcp = obj.fr.mcp, 
       obj.pqfr  =obj.pqfr,
       obj.pqfr.scad  =obj.pqfr.scad,
       obj.pqfr.mcp  =obj.pqfr.mcp, 
       obj.lasso = obj.lasso,
       obj.alasso = obj.alasso
  )
}


###########################################################################
# Part I: main functions
# Part I 2) RDA.pred: evaluate the prediction error (PE) off all methods of the paper
#
# Input:
#   x:  n x p design matrix
#   y: the response vector
#   tau.vec: a vector of the quantile levels
#   seed.vec: a vector of random seeds
#
# Output: 
#   a summary table of PE for all methods considered in the paper


RDA.pred <- function(x, y, tau.vec, seed.vec){
  
  B <- length(seed.vec)
  n <- length(y)
  p <- ncol(x)
  n.tau <- length(tau.vec)
  
  pred.sc <- pred.fr <- matrix(NA, nrow=B, ncol=n.tau)
  pred.lasso <- pred.scad <- pred.mcp <- matrix(NA, nrow=B, ncol=n.tau)
  pred.pqfr <- pred.pq <- matrix(NA, nrow=B, ncol=n.tau)

  nz.sc <- nz.fr <- matrix(NA, nrow=B, ncol=n.tau)
  nz.lasso <- nz.scad <- nz.mcp <- matrix(NA, nrow=B, ncol=n.tau)
  nz.pqfr <- nz.pq <- matrix(NA, nrow=B, ncol=n.tau)
  
  
  eval.pred <- function(beta.est, x.test, y.test, tau){
    mean(check(y.test - cbind(1,x.test)%*%beta.est, tau))
  }
  
  for(b in 1:B){
    
    set.seed(seed.vec[b])
    
    ind.train <- sort(sample(1:n, n/2, replace=FALSE), decreasing=FALSE)
    ind.test <- setdiff(1:n, ind.train)
    x.train <- x[ind.train,]; x.test <- x[ind.test,]
    y.train <- y[ind.train]; y.test <- y[ind.test]

    for(i in 1:length(tau.vec)){
      obj <- RDA(x.train, y.train, tau.vec[i])

      pred.pq[b,i] <- eval.pred(obj$beta.pq, x.test, y.test, tau.vec[i])
      pred.lasso[b,i] <- eval.pred(obj$beta.lasso, x.test, y.test, tau.vec[i])
      pred.scad[b,i] <- eval.pred(obj$beta.scad, x.test, y.test, tau.vec[i])
      pred.mcp[b,i] <- eval.pred(obj$beta.mcp, x.test, y.test, tau.vec[i])
      pred.sc[b,i] <- eval.pred(obj$beta.sc, x.test, y.test, tau.vec[i])
      pred.fr[b,i] <- eval.pred(obj$beta.fr, x.test, y.test, tau.vec[i])
      pred.pqfr[b,i] <- eval.pred(obj$beta.pqfr, x.test, y.test, tau.vec[i])

      nz.pq[b,i] <- sum(obj$beta.pq[-1]!=0)
      nz.lasso[b,i] <- sum(obj$beta.lasso[-1]!=0)
      nz.scad[b,i] <- sum(obj$beta.scad[-1]!=0)
      nz.mcp[b,i] <- sum(obj$beta.mcp[-1]!=0)
      nz.sc[b,i] <- sum(obj$beta.sc[-1]!=0)
      nz.fr[b,i] <- sum(obj$beta.fr[-1]!=0)
      nz.pqfr[b,i] <- sum(obj$beta.pqfr[-1]!=0)
    }
    
    cat(b, '/')
  }

  
  pred.mean <- round(rbind(
        colMeans(pred.pq), 
        colMeans(pred.lasso),
        colMeans(pred.scad),
        colMeans(pred.mcp),
        colMeans(pred.sc),
        colMeans(pred.fr),
        colMeans(pred.pqfr)
        ), 3)
  pred.sd <- round(rbind(
    apply(pred.pq, 2, sd), 
    apply(pred.lasso, 2, sd),
    apply(pred.scad, 2, sd),
    apply(pred.mcp, 2, sd),
    apply(pred.sc, 2, sd),
    apply(pred.fr, 2, sd),
    apply(pred.pqfr, 2, sd)
  ), 3)
  
  nz.mean <- round(rbind(
    colMeans(nz.pq), 
    colMeans(nz.lasso),
    colMeans(nz.scad),
    colMeans(nz.mcp),
    colMeans(nz.sc),
    colMeans(nz.fr),
    colMeans(nz.pqfr)
  ), 3)
  
  nz.sd <- round(rbind(
    apply(nz.pq, 2, sd), 
    apply(nz.lasso, 2, sd),
    apply(nz.scad, 2, sd),
    apply(nz.mcp, 2, sd),
    apply(nz.sc, 2, sd),
    apply(nz.fr, 2, sd),
    apply(nz.pqfr, 2, sd)
  ), 3)

  out <- data.frame(
    pred.mean[,1], pred.sd[,1], nz.mean[,1], nz.sd[,1], 
    pred.mean[,2], pred.sd[,2], nz.mean[,2], nz.sd[,2],
    pred.mean[,3], pred.sd[,3], nz.mean[,3], nz.sd[,3])
  
  rownames(out) <- c("pq", "lasso", "scad", "mcp", 
                    "sc", "fr", "pqfr")
  
  out
  
}


###########################################################################
# Part II: sub functions
# Part II 1) rq.margin: main function for calculating the (conditional) marginal utility
#
# Input:
#   z: n x p design matrix
#   y: the response 
#   tau: the quantile level
#   ind.act: the conditioning set
#   ind.cand: the candidate set
#   intercept: whether or not to include the intercept in the model (TRUE or FALSE)
#
# Output: 
#   score: the score for the marignal utility (condition on the ind.act)


rq.margin <- function(z, y, tau, ind.act=NULL, ind.cand=NULL, intercept=TRUE, method="br"){
  
  z <- as.matrix(z)
  n <- nrow(z)
  p <- ncol(z)
  if(is.null(ind.cand)){
    ind.cand <- setdiff(1:p, ind.act)  
  }
  n.ind.act <- length(ind.act)
  n.ind.cand <- length(ind.cand)
  score.m <- rep(NA, p)
  
  ### Marginal regression
  for(i in 1:n.ind.cand){
    ind <- sort(c(as.numeric(ind.act), as.numeric(ind.cand[i])))
    if(intercept){
      obj <- rq(y ~ as.matrix(z[,ind]), tau=tau, method=method)  
    } else {
      obj <- rq(y ~ as.matrix(z[,ind]) - 1, tau=tau, method=method)  
    }
    score.m[ind.cand[i]] <- mean(check(obj$residuals, tau))
    if(is.infinite(abs(score.m[ind.cand[i]]))){
      score.m[ind.cand[i]] <- NA
    }
  }
  
  return(score=score.m)
}
###########################################################################
# Part II: sub functions
# Part II 2) rq.greedy: main function for generating the greedy path
#			 rq.greedy.h: main function for generating the hybrid path
#
# Input:
#   x: n x p design matrix
#   y: the response 
#   tau: the quantile level
#   screen.method: choose the method of greedy algorithm (SC, gSC, FR)
#   stop.method: top the greedy algorithm by LIC ( or set ``none'' to stop until k hits max.iter)
#   intercept: whether or not to include the intercept in the model (TRUE or FALSE)
#   xi: the tuning parameter for LIC (xi in the paper)
#   m: the parameter for gSC (m0 in the paper)
#   K: stop the algorithm when LIC is satisfies for K times (T in the paper)
#   max.iter: the maximum parameter (Kn in the paper)
# Output: 
#   beta: the solution path generated by the greedy algorithm
#   loss: the loss corresponds to the solution path beta
#   ind.act: the index set selected by the greedy algorithm
#   LIC: a sequence of LIC values corresponds to the greedy path
#   HQIC: a sequence of HQIC values corresponds to the greedy path
#   runtime: computing time


rq.greedy <- function(x, y, tau, screen.method, stop.method="none", xi=1, m=50, K=3, max.iter=30, method="br", c0=1){

  x <- as.matrix(x)
  n <- nrow(x); p <- ncol(x)
  obj <- rq(y ~ 1, tau=tau)
  Res <- obj$residuals
  
  time.start <- Sys.time()

  ### initial setting
  info.vec <- rep(NA, max.iter)
  beta.path <- matrix(0, nrow=p+1, ncol=max.iter)
  rownames(beta.path) <- c("intercept", 1:p)
  ind.LIC <- ind.QIC <- ind.act <- NULL
  ind.list <- vector("list", max.iter)
  ind.cand <- setdiff(1:p, ind.act)  
  LIC.vec <- HQIC.vec <- df <- loss.vec <- rep(NA, max.iter)
  k0 <- 0

  for(i in 1:max.iter){

    ### variable screening
    if(screen.method=="SC"){
      score <- rq.margin(x, Res, tau=tau, ind.act=NULL, intercept=FALSE, method=method)
    } else if (screen.method=="FR"){
      score <- rq.margin(x, y, tau=tau, ind.act=ind.act, method=method)
    } else if (screen.method=="PQU.fr"){
      score <- -Screen_Mar(x, Res, tt=tau, 1)
    } else if (screen.method=="CGA"){
      grad <- crossprod(x, tau - ifelse(Res<=0, 1, 0))/n
      score <- -abs(grad)
    } else if (screen.method=="SC+FR"){
      s0 <- rq.margin(x, Res, tau=tau, ind.act=NULL, ind.cand=ind.cand, intercept=FALSE, method=method)
      s0[ind.act] <- Inf
      ind <- which((p-rank(-s0, ties.method="max")+1)<=m)
      score <- rq.margin(x, y, tau=tau, ind.act=ind.act, ind.cand=ind, method=method)
    } else if (screen.method=="CGA+FR"){
      grad <- crossprod(x, tau - ifelse(Res<=0, 1, 0))/n
      s0 <- -abs(grad); s0[ind.act] <- Inf
      ind <- which((p-rank(-s0, ties.method="max")+1)<=m)
      score <- rq.margin(x, y, tau=tau, ind.act=ind.act, ind.cand=ind, method=method)
    }
    score[ind.act] <- Inf
    ind.act <- c(ind.act, which.min(score))
    ind.cand <- setdiff(1:p, ind.act)  
    df[i] <- length(ind.act)

    ### refit MLE (with intercept)
    obj <- rq(y ~ as.matrix(x[,ind.act]), tau=tau, method=method)
    Res <- obj$residuals
    loss.vec[i] <- mean(check(Res, tau))
    beta.path[c(1,ind.act+1), i] <- obj$coefficients
    
    if(is.nan(loss.vec[i])|is.infinite(abs(loss.vec[i]))){
      i <- i-1;
      ind.act <- ind.act[1:i]
      break;
    }
    
    ### recording information criterion
    LIC.vec[i] <- loss.vec[i] + (df[i]^2)*xi*log(p)/n
    HQIC.vec[i] <- log(loss.vec[i]*n) + c0*df[i]*log(n)*log(log(p))/n/2
    
    ### stopping criterion
    if(i>1){
      if(stop.method=="LIC"){
        if(loss.vec[i-1]-loss.vec[i] < xi*(i-1)*log(p)/n){
          k0 <- k0 + 1
        } else {
          k0 <- 0
        }
      } 
    }

    # cat("i=", i, "; loss=", loss.vec[i], '\n')
    # print(ind.act)

    if(k0==K) break;
    if(i==max.iter) break;
  }
  
  # beta.path <- sweep(beta.path, 1, normx, "/")
  
  time.end <- Sys.time()
  runtime <- difftime(time.end, time.start, units="secs")
  
  list(beta=beta.path[,1:i], 
       loss=loss.vec[1:i],
       ind.act=ind.act,
       LIC=LIC.vec[1:i],
       HQIC=HQIC.vec[1:i],
       runtime=runtime)
}


rq.greedy.h <- function(x, y, tau, stop.method = "none", 
                        xi = 1, m.vec = c(1,25,50), K = 3, max.iter = 30, method = "br", 
                        c0 = 1){
  
  x <- as.matrix(x)
  n <- nrow(x); p <- ncol(x)
  m <- ceiling(n/log(n))
  m.vec <- c(1, m, p)
  
  obj.list <- vector("list", length(m.vec))
  for(i in 1:(length(m.vec))){
    if(i==1){
      obj.list[[i]] <- rq.greedy(x, y, tau=tau, screen.method="SC", stop.method=stop.method, xi=xi, K=K, max.iter=max.iter, method=method, c0=c0)
    } else if(i==p){
      obj.list[[i]] <- rq.greedy(x, y, tau=tau, screen.method="FR", stop.method=stop.method, xi=xi, K=K, max.iter=max.iter, method=method, c0=c0) 
    } else {
      obj.list[[i]] <- rq.greedy(x, y, tau=tau, screen.method="SC+FR", stop.method=stop.method, xi=xi, K=K, max.iter=max.iter, method=method, c0=c0)
    }
  }
  
  ind.act <- NULL; runtime <- 0
  for(i in 1:length(m.vec)){
    ind.act <- union(ind.act, obj.list[[i]]$ind.act)
    runtime <- runtime + obj.list[[i]]$runtime
  }
  ind.act <- sort(ind.act)
  
  obj <- rq(y ~ as.matrix(x[, ind.act]), tau=tau)
  beta <- rep(0, p+1); beta[c(1,ind.act+1)] <- coef(obj)
  loss <- mean(check(obj$residuals, tau))
  
  list(beta=beta, loss=loss, ind.act=ind.act, runtime=runtime)
}


###########################################################################
# Part II: sub functions
# Part II 3) QCIDpen2: to implement SCAD and MCP for quantile regression models
#              ALASSO: to implement adaptive lasso for quantile regression models
#
# Input:
#   x: n x p design matrix
#   y: the response 
#   tau: the quantile level
#   beta.init: the initial value
#   intercept: whether or not to include the intercept in the model (TRUE or FALSE)
# Output: 
#   beta: the solution path
#   lambda: a sequence of lambda
#   loss: the loss corresponds to the solution path beta
#   HQIC: a sequence of HQIC values corresponds to the greedy path
#   beta.hqic: the regression estimates corresponding to the minimum HQIC value
#   ind.hqic: the nonzero indices of beta.hqic
#   runtime: computing time


QICD.pen2 <- function(x, y, tau, penalty, ind=NULL, beta.init=NULL, intercept=TRUE, c0=1){
  
  n <- nrow(x)
  p <- ncol(x)
  
  
  nlambda <- 100
  a0 <- sqrt(log(p)/n)
  lambda <- a0*exp(seq(log(100), log(1e-3), len=nlambda))
  beta.path <- matrix(0, nrow=p+1, ncol=nlambda)
  HQIC.vec <- loss.vec <- rep(NA, nlambda)
  if(is.null(beta.init)) beta.init <- rep(0, p+1)
  if(!is.null(ind)) beta.init <- rep(0,length(ind)+1)
  
  time.start <- Sys.time()
  
  for(i in 1:nlambda){
    
    if(!is.null(ind)){
      xs <- as.matrix(x[,ind])
      beta.path[c(1,ind+1),i] <- beta.init <- QICD(y, xs, tau=tau, lambda=lambda[i], intercept=intercept, penalty=penalty, initial_beta=beta.init)
    } else {
      beta.init <- beta.path[,i] <- QICD(y, x, tau=tau, lambda=lambda[i], intercept=intercept, penalty=penalty, initial_beta=beta.init)  
    }
    beta.path[abs(beta.path[,i])<1e-5,i] <- 0
    df <- sum(beta.path[,i]!=0)
    if(df > n) break;
    loss.vec[i] <- mean(check(y-cbind(1,x)%*%beta.path[,i], tau))
    HQIC.vec[i] <- log(loss.vec[i]*n) + c0*df*log(n)*log(p)/n/2
    # cat(i, '/')
  }
  time.end <- Sys.time()
  runtime <- difftime(time.end, time.start, units="secs")
  
  beta.hqic <- beta.path[, which.min(HQIC.vec)]
  ind.hqic <- which((beta.hqic[-1])!=0)
  
  list(beta=beta.path, lambda=lambda, loss=loss.vec, HQIC=HQIC.vec, 
       beta.hqic=beta.hqic, ind.hqic=ind.hqic, runtime=runtime)
}

ALASSO <- function(x, y, tau, ind=NULL, method="quantile", penalty.factor, c0=1){
  
  n <- length(y)
  p <- ncol(x)
  
  time.start <- Sys.time()
  if(!is.null(ind)){
    xs <- x[,ind]
    pf <- penalty.factor[ind]
    obj <- hqreg_raw(xs, y, method="quantile", tau=tau, penalty.factor=pf, lambda.min=1e-4)
    lambda <- obj$lambda
    beta.path <- matrix(0, nrow=p+1, ncol=length(lambda))
    beta.path[c(1,ind+1),] <- obj$beta
  } else {
    obj <- hqreg_raw(x, y, method="quantile", tau=tau, penalty.factor=penalty.factor, lambda.min=1e-4)
    lambda <- obj$lambda
    beta.path <- obj$beta
  }
  
  loss.vec <- apply(cbind(1,x)%*%beta.path, 2, function(x) sum(check(y-x, tau)))
  df <- colSums(beta.path!=0)
  HQIC.vec <- log(loss.vec) + c0*df*log(n)*log(p)/n/2
  beta.hqic <- beta.path[, which.min(HQIC.vec)]
  ind.hqic <- which(beta.hqic[-1]!=0)
  time.end <- Sys.time()
  runtime <- difftime(time.end, time.start, units="secs")
  
  
  list(beta=beta.path, lambda=lambda, loss=loss.vec, HQIC=HQIC.vec, 
       beta.hqic=beta.hqic, ind.hqic=ind.hqic, runtime=runtime)
}

###########################################################################
# Part II: sub functions
# Part II 4) Screen_Mar: to implement the marginal screening method as in Biometrika 2014 
#
# Input:
#   x is the covariate matrix 
#   y is the response 
#   tt is the quantile of interest between 0 and 1
#   mS =  n/log(n)   --- model size 
# Output: 
#   gives the indexes of covariates that are non-zero or relevant -- indiacted by 1
#
# This code is downloaded from Yinfei Kong's website
# (https://sites.google.com/site/yinfeikong/research)


Screen_Mar <- function(x, y, tt, mS)
{
  
  # x is the covariate matrix 
  # y is the response 
  # tt is the quantile of interest between 0 and 1
  # mS =  n/log(n)   --- model size 
  
  #output: gives the indexes of covariates that are non-zero or relevant -- indiacted by 1
  
  
  Covarian_I <- function(x, gnn)
  {
    
    # the covariates -- n by p matrix   
    # gnn =  the number of grid points 
    # Outout = an n by gnn by p array -- this is actually an n by gnn matrix for each covariate
    
    n <- dim(x)[1]
    p <- dim(x)[2]
    HH <- array(0, c(n, gnn, p))
    
    for (k in 1:p)   
    {
      #k --- no of features
      z <- x[, k]
      zmi <- quantile(z, 0.05)
      zma <- quantile(z, 0.95)
      z.g <- seq(zmi, zma, length.out=gnn) # fixed grid to help reduce computational time,,
      #z0 <- matrix(rep(z, n), n, n)
      z0 <- matrix(rep(z, gnn), n, gnn)
      #z1 <- matrix(rep(z, each=n), n, n)
      z1 <- matrix(rep(z.g, each=n), n, gnn)
      HH[,,k] <- (z0 < z1) * 1
      
    }
    return(HH)
  }  
  
  
  n <- dim(x)[1]
  p <- dim(x)[2]
  Est <- matrix(0, 3, p)        # 3 is for the number of Cn values
  
  
  
  gnn <- 100  # grid size 
  HH <- Covarian_I(x, gnn)
  
  kk <- 1:p  
  NonIndex0 <- c() 
  Q_t <- quantile(y, tt)  
  AA <-  tt - (y < Q_t) * 1
  AA1 <- matrix(rep(AA, gnn), n, gnn)
  
  
  pss <- ceiling(p/2)
  HH1 <- HH[,,1:pss]
  HH2 <- HH[,,-(1:pss)]
  
  
  
  dis1.m <- log(apply(do.call(cbind,lapply(seq_len(dim(HH1)[3]),
                                           function(i) apply(HH1[,,i]* AA1, 2, mean)^2 )),  2, mean))
  
  dis2.m <- log(apply(do.call(cbind,lapply(seq_len(dim(HH2)[3]),
                                           function(i) apply(HH2[,,i]* AA1, 2, mean)^2 )),  2, mean))
  
  #dis.m <- log(apply(do.call(cbind,lapply(seq_len(dim(HH)[3]),
  #  function(i) apply(HH[,,i]* AA1, 2, mean)^2 )),  2, mean))
  
  
  dis.m <- c(dis1.m, dis2.m)
  dis.ms <- sort(dis.m, index.return=TRUE, decreasing=TRUE)
  # Selected variables (their indexes) using marginal quantiles
  
  Est <- dis.ms$ix[1:mS]
  kk0 <- rep(0, p)
  kk0[Est] <- 1
  
  return(kk0)
  
}
